import ConfigParser
import json
import logging
import socket
import StringIO
import struct
import time

logger = logging.getLogger(__name__)


class ZabbixMetric(object):
    """
    Create structure contains metric for zabbix server.

    Attributes:
    host (str): Hostname as it displays in Zabbix.
    key (str):  Key by which you will identify this metric.
    value (*):  Key value.
    clock (int): A timestamp in seconds. If None, then it will current time

    Example:
    >> ZabbixMetric('localhost', 'cpu[usage]', 20)
    """

    def __init__(self, host, key, value, clock=None):
        self.host = str(host)
        self.key = str(key)
        self.value = str(value)
        self.clock = clock if clock else str(int(time.time()))

    def __repr__(self):
        """
        Represent detailed ZabbixMetric view.
        """

        result = json.dumps(self.__dict__)
        logger.debug('%s: %s', self.__class__.__name__, result)

        return result


class ZabbixSender(object):
    """
    Send ZabbixMetrics to Zabbix server, like zabbix_sender util.
    Implement zabbix trapper protocol.

    Attributes:
      zabbix_server (str):  IP address of Zabbix server. Default: 127.0.0.1

      zabbix_port (int):    Port of Zabbix server. Default: 10051

      use_config (str):     Specify config file from which zabbix_server
                            and zabbix_port will be loaded.

                            If True then use default config:
                              /etc/zabbix/zabbix_agentd.conf
    """

    def __init__(self, zabbix_server='127.0.0.1', zabbix_port=10051, use_config=None):

        self.cn = self.__class__.__name__

        if use_config:
            self.zabbix_uri = self.__load_from_config(use_config)
        else:
            self.zabbix_uri = [(zabbix_server, zabbix_port)]

        logger.debug('%s(%s)', self.cn, self.zabbix_uri)

    @classmethod
    def __load_from_config(cls, config_file):
        """
        Load zabbix server ip address and port from zabbix agent file

        If Server or Port variable won't be found in the file, they will be
        set up from defaults: 127.0.0.1:10051
        """

        if config_file and isinstance(config_file, bool):
            config_file = '/etc/zabbix/zabbix_agentd.conf'

        try:
            #  This is workaround for config wile without sections
            with open(config_file, 'r') as f:
                config_file_data = "[root]\n" + f.read()
        except:
            exit()

        config_file_fp = StringIO.StringIO(config_file_data)
        config = ConfigParser.RawConfigParser({'Server': '127.0.0.1', 'Port': 10051})
        config.readfp(config_file_fp)
        zabbix_server = config.get('root', 'Server')
        zabbix_port = config.get('root', 'Port')
        zabbix_server_list = [server.strip() for server in zabbix_server.split(',')]

        result = [(server, zabbix_port) for server in zabbix_server_list]

        return result

    @classmethod
    def __receive(cls, sock, count):
        """
        Reads socket to receive data from zabbix server.

        Attributes:
          socket (socket):  Socket object from which data should be read
          count (int):  Amount of data that should be read from socket, in bytes.
        """

        buf = ''

        while len(buf) < count:
            chunk = sock.recv(count - len(buf))
            if not chunk:
                break
            buf += chunk

        return buf

    def __create_messages(self, metrics_array):
        """
        Create a list of zabbix messages, from a list of ZabbixMetrics.

        Attributes:
          metrics_array (list of ZabbixMetrics): List of ZabbixMetrics

        Returns:
          list of str: List of zabbix messages
        """

        metrics = []

        # Fill the array of messages
        for m in metrics_array:
            metrics.append(str(m))

        logger.debug('%s.__create_messages: %s', self.cn, metrics)

        return metrics

    def __create_request(self, messages):
        """
        Create a formatted request to zabbix from a list of messages.

        Attributes:
          messages (list of str): List zabbix messages

        Returns:
          str: Formatted request to zabbix
        """

        request = '{"request":"sender data","data":[%s]}' % ','.join(messages)
        logger.debug('%s.__create_request: %s', self.cn, request)

        return request

    def __create_packet(self, request):
        """
        Create a formatted packet from a request.

        Attributes:
          request (str): Request string to zabbix

        Returns:
          str: Packet string to zabbix
        """

        data_len = struct.pack('<Q', len(request))
        packet = 'ZBXD\x01' + data_len + request
        logger.debug('%s.__create_packet (str): %s', self.cn, packet)
        logger.debug('%s.__create_packet (hex): %s', self.cn,
                     ':'.join(x.encode('hex') for x in packet))

        return packet

    def __get_response(self, connection):
        """
        Get response from zabbix server, reads from self.socket.

        Returns:
          str: JSON response from zabbix server or False in case of some errors
        """

        response_header = self.__receive(connection, 13)
        logger.debug('%s.__get_response.response_header: %s', self.cn, response_header)

        if not response_header.startswith('ZBXD\x01') or len(response_header) != 13:
            logger.debug('%s.__get_response: Wrong zabbix response', self.cn)
            result = False
        else:
            response_len = struct.unpack('<Q', response_header[5:])[0]

            response_body = connection.recv(response_len)

            result = json.loads(response_body)
            logger.debug('%s.__get_response: %s', self.cn, result)

        try:
            connection.close()
        except:
            pass

        return result

    def send(self, metrics):
        """
        Send the metrics to zabbix server.

        Attributes:
          metrics (list of ZabbixMetrics): List of metrics that will be send to Zabbix

        Returns:
          bool: True if sent successful, False if was an error.
        """

        result = None

        messages = self.__create_messages(metrics)
        request = self.__create_request(messages)
        packet = self.__create_packet(request)

        for host_addr in self.zabbix_uri:
            logger.debug('%s.send(%s): connecting', self.cn, host_addr)

            # create socket object
            connection = socket.socket()

            # server and port must be tuple
            connection.connect(host_addr)

            try:
                connection.sendall(packet)
            except Exception as e:
                logger.debug("%s.send: Error while sending the data to zabbix\nERROR:%s", self.cn, e)
                connection.close()
                exit()

            # socket will be closed in self._get_response()
            response = self.__get_response(connection)
            logger.debug('%s.send(%s): %s', self.cn, host_addr, response)

            if response and response.get('response') == 'success':
                result = True
            else:
                logger.debug('%s.send: Got error from zabbix => %s}', self.cn, response)
                raise Exception('Zabbix Server ({0}) reject packet.'.format(host_addr))

        return result



#!/usr/bin/env python
from setuptools import setup

setup(name='py-zabbix',
      version='0.5.3',
      description='Python modules for work with zabbix.',
      url='https://github.com/blacked/py-zabbix',
      author='Alexey Dubkov',
      author_email='alexey.dubkov@gmail.com',
      packages=['zabbix']
     )



import json
import logging
import re
import ssl
import urllib2

class _NullHandler(logging.Handler):
  """
  Null Handler class for Logger
  """

  def emit(self, record):
    pass

logger = logging.getLogger(__name__)
logger.addHandler(_NullHandler())

class ZabbixAPIException(Exception):
  """
  ZabbixAPI exception class

  code list:
  -32602 - Invalid params (eg already exists)
  -32500 - no permissions
  """
  pass

class ZabbixAPIObjectClass(object):
  """
  ZabbixAPI Object class
  """

  def __init__(self, name, parent):
    self.name = name
    self.parent = parent

  def __getattr__(self, attr):
    """
    Dynamically create a method (ie: get)
    """

    def fn(*args, **kwargs):
      if args and kwargs:
        raise TypeError("Found both args and kwargs")

      logger.debug(attr)

      return self.parent.do_request(
          '{0}.{1}'.format(self.name, attr),
          args or kwargs
      )['result']

    return fn

class ZabbixAPI(object):
  """
  ZabbixAPI class, implement interface to zabbix api

  Examples:
    z = ZabbixAPI('https://zabbix.server', user='admin', password='zabbix')

    # Get API Version
    z.api_info.version()
    >> u'2.2.1'
    # ir
    z.do_request('apiinfo.version')
    >> {u'jsonrpc': u'2.0', u'result': u'2.2.1', u'id': u'1'}

    # Get all disabled hosts
    z.host.getobjects(status=1)
    # or
    z.do_request('host.getobjects', {'status':1})

  """

  def __init__(self, url='https://localhost/zabbix',
               use_authenticate=False, user='admin', password='zabbix'):
    self.use_authenticate = use_authenticate
    self.auth = None
    self.url = url + '/api_jsonrpc.php'
    self.__login(user, password)
    logger.debug("JSON-PRC Server: %s", self.url)

  def __getattr__(self, attr):
    """
    Dynamically create an object class (ie: host)
    """
    return ZabbixAPIObjectClass(attr, self)

  def __login(self, user='', password=''):
    """
    Do login to zabbix server

    Attributes:
      user (str):     Zabbix user
      password (str): Zabbix user password
    """
    logger.debug("ZabbixAPI.login({0},{1})".format(user, password))

    self.auth = None

    if self.use_authenticate:
      self.auth = self.user.authenticate(user=user, password=password)
    else:
      self.auth = self.user.login(user=user, password=password)

  def api_version(self):
    """
    Return version of Zabbix API
    """
    return self.apiinfo.version()

  def do_request(self, method, params=None):
    """
    Make request to Zabbix API

    Attributes:
      method (str): Any of ZabbixAPI method, like: apiinfo.version
      params (str): Methods parameters

    Examples:
      z = ZabbixAPI()
      apiinfo = z.do_request('apiinfo.version')
    """
    request_json = {
      'jsonrpc':'2.0',
      'method': method,
      'params': params or {},
      'auth': self.auth,
      'id': '1',
    }

    # Create default context to skip SSL cert verification.
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    logger.debug('urllib2.Request({0}, {1})'.format(self.url,json.dumps(request_json)))
    req = urllib2.Request(self.url, json.dumps(request_json))
    req.get_method = lambda: 'POST'
    req.add_header('Content-Type', 'application/json-rpc')

    try:
      res = urllib2.urlopen(req, context=ctx)
      response_json = json.load(res)
    except ValueError:
      raise ZabbixAPIException("Unable to parse json: %" % res)

    logger.debug("Response Body: %s" % json.dumps(response_json, indent=4,
      separators=(',', ': ')))

    if 'error' in response_json:
      msg = "Error {code}: {message}, {data} while sending {json}".format(
        code=response_json['error']['code'],
        message=response_json['error']['message'],
        data=response_json['error']['data'],
        json=str(request_json)
      )
      raise ZabbixAPIException(msg, response_json['error']['code'])

    return response_json

  def get_id(self, item_type, item=None, with_id=False, hostid=None, **args):
    """
    Return id or ids of zabbix objects.

    Attributes:
      item_type   (str): Type of zabbix object
      item        (str): Name of zabbix object.
                         If None - will return list of all object in the scope.
      with_id    (bool): Return values will be in zabbix format.
                         Examlpe: {'itemid: 128}
      hostid      (int): Specify id of host for special cases
      templateid  (int): Specify scope to specific template
      app_name    (str): Specify scope to specific template

    """
    result = None

    type_ = '{item_type}.get'.format(item_type=item_type)

    item_filter_name = {
      'trigger': 'description',
      'triggerprototype': 'description',
      'mediatype': 'description',
      'user': 'alias',
      'usermacro': 'macro',
    }

    item_id_name = {
      'usermacro': 'hostmacro',
      'usergroup': 'usrgrp',
      'hostgroup': 'group',
      'discoveryrule': 'item',
      'graphprototype': 'graph',
      'itemprototype': 'item',
      'triggerprototype': 'trigger',
    }

    filter_ = { 'filter': { item_filter_name.get(item_type, 'name'): item }, 'output':'extend' }
    if hostid:
      filter_['filter'].update({ 'hostid': hostid })

    if args.get('templateid'):
      filter_['templateids'] = args['templateid']
    if args.get('app_name'):
      filter_['application'] = args['app_name']

    logger.debug('do_request( "{type}", {filter} )'.format(type=type_, filter=filter_))
    response = self.do_request(type_, filter_)['result']

    if response:
      item_id = '{item}id'.format(item=item_id_name.get(item_type, item_type))
      result = []
      for obj in response:
        if with_id:
          result.append({ item_id: int(obj.get(item_id)) })
        else:
          result.append(int(obj.get(item_id)))

      list_types = (list, type(None))
      if not isinstance(item, list_types):
        result = result[0]

    return result
